;; -*- mode: scheme; coding: utf-8 -*-
;; Copyright © 2020 Göran Weinholt <goran@weinholt.se>
;; SPDX-License-Identifier: MIT
#!r6rs

;;; Serializer and deserializer for the D-Bus wire format

;; Wherein our heroes did not care for ASN.1

(library (d-bus protocol wire)
  (export
    d-bus-put-message
    d-bus-unmarshal-message)
  (import
    (rnrs)
    (d-bus compat match)
    (d-bus protocol signatures)
    (d-bus protocol utils))

#|

  This representation is used for D-Bus values:

  * All basic types have their natural Scheme equivalents.
  * Strings, object paths and signatures are all strings.
  * Structs are represented by lists.
  * Dict entries are (key . value) pairs.
  * Variants are ("signature" . value).
  * Arrays are vectors.

|#

(define debug #f)

;; Deserialize a message from a bytevector.
(define (d-bus-unmarshal-message bv offset endian sig)
  (define who 'd-bus-unmarshal-message)
  (define (fail x)
    (assertion-violation who "Invalid signature" sig x))
  (define (unmarshal-datum bv offset endian type)
    (when debug
      (write `(unmarshal-datum bv ,(number->string offset 16) ,endian ,type))
      (newline))
    (match type
      ['BYTE
       (values (fx+ offset 1) (bytevector-u8-ref bv offset))]
      ['BOOLEAN
       (let ((offset^ (fxalign offset 4)))
         (values (fx+ offset^ 4)
                 (case (bytevector-u32-native-ref bv offset^)
                   [(0) #f]
                   [(1) #t]
                   [else (fail type)])))]
      ['INT16
       (let ((offset (fxalign offset 2)))
         (values (fx+ offset 2) (bytevector-s16-ref bv offset endian)))]
      ['UINT16
       (let ((offset (fxalign offset 2)))
         (values (fx+ offset 2) (bytevector-u16-ref bv offset endian)))]
      ['INT32
       (let ((offset (fxalign offset 4)))
         (values (fx+ offset 4) (bytevector-s32-ref bv offset endian)))]
      [(or 'UINT32 'UNIX_FD)
       (let ((offset (fxalign offset 4)))
         (values (fx+ offset 4) (bytevector-u32-ref bv offset endian)))]
      ['INT64
       (let ((offset (fxalign offset 8)))
         (values (fx+ offset 8) (bytevector-s64-ref bv offset endian)))]
      ['UINT64
       (let ((offset (fxalign offset 8)))
         (values (fx+ offset 8) (bytevector-u64-ref bv offset endian)))]
      ['DOUBLE
       (let ((offset (fxalign offset 8)))
         (values (fx+ offset 8) (bytevector-ieee-double-ref bv offset endian)))]
      [(or 'STRING 'OBJECT_PATH)
       (let* ((offset (fxalign offset 4))
              (len (bytevector-u32-ref bv offset endian))
              (str (make-bytevector len)))
         (bytevector-copy! bv (fx+ offset 4) str 0 (bytevector-length str))
         (values (fx+ offset (fx+ len (fx+ 4 1))) (utf8->string str)))]

      ['SIGNATURE
       (let* ((len (bytevector-u8-ref bv offset))
              (str (make-bytevector len)))
         (bytevector-copy! bv (fx+ offset 1) str 0 (bytevector-length str))
         (values (fx+ offset (fx+ len (fx+ 1 1))) (utf8->string str)))]

      [('STRUCT . type*)
       (let ((offset (fxalign offset 8)))
         (d-bus-unmarshal-message bv offset endian `(message ,@type*)))]

      [('DICT_ENTRY . type*)
       (let ((offset (fxalign offset 8)))
         (let-values ([(offset vals)
                       (d-bus-unmarshal-message bv offset endian `(message ,@type*))])
           (match vals
             [(k v)
              (values offset `(,k . ,v))])))]

      ['VARIANT
       (let* ((len (bytevector-u8-ref bv offset))
              (str (make-bytevector len)))
         (bytevector-copy! bv (fx+ offset 1) str 0 (bytevector-length str))
         (let ((sig (utf8->string str)))
           (let ((sig^ (parse-type-signature sig)))
             (let ((offset (fx+ offset (fx+ len (fx+ 1 1)))))
               (let-values ([(offset vals)
                             (d-bus-unmarshal-message bv offset endian sig^)])
                 (values offset (cons (cadr sig^) (car vals))))))))]

      [('ARRAY atype)
       (let* ((offset (fxalign offset 4))
              (len (bytevector-u32-ref bv offset endian))
              (offset (fx+ offset 4))
              (end (fx+ offset len))
              (type `(message ,atype)))
         (let lp ((offset offset) (vals '()))
           (cond ((fx=? offset end)
                  (values offset (list->vector (reverse vals))))
                 ((fx>? offset end)
                  (assertion-violation who "Array length incorrect" sig offset end))
                 (else
                  (let-values ([(offset vals^) (d-bus-unmarshal-message bv offset endian type)])
                    (let ((vals (append vals^ vals)))
                      (lp offset vals)))))))]

      [_ (fail type)]))

  (when debug
    (write `(d-bus-unmarshal-message bv ,(number->string offset 16) ,endian ,sig))
    (newline))
  (match sig
    [('message type* ...)
     (let lp ((type* type*) (offset offset) (vals '()))
       (match type*
         [()
          (values offset (reverse vals))]
         [(type . type*^)
          (let-values ([(offset^ val) (unmarshal-datum bv offset endian type)])
            (when debug
              (write (list 'offset offset '=> val))
              (newline))
            (lp type*^ offset^ (cons val vals)))]))]
    [_ (fail sig)]))

;; Serialize a message to a port using native endianness
(define (d-bus-put-message port offset post-align sig data*)
  (define who 'd-bus-put-message)
  (define (align-port port offset alignment)
    (let ((padding (fx- (fxalign offset alignment) offset)))
      (put-bytevector port #vu8(0 0 0 0 0 0 0 0) 0 padding)
      (fx+ offset padding)))
  (define (fail x)
    (assertion-violation who "Invalid signature" sig x))
  (define buffer (make-bytevector 8))
  (define (marshal-datum! buffer offset type datum)
    (match type
      ;; Basic types
      ['BYTE        (bytevector-u8-set! buffer 0 datum)
                    (values 1 1 #f)]
      ['BOOLEAN     (bytevector-u32-native-set! buffer 0 (if datum 1 0))
                    (values 4 4 #f)]
      ['INT16       (bytevector-s16-native-set! buffer 0 datum)
                    (values 2 2 #f)]
      ['UINT16      (bytevector-u16-native-set! buffer 0 datum)
                    (values 2 2 #f)]
      ['INT32       (bytevector-s32-native-set! buffer 0 datum)
                    (values 4 4 #f)]
      [(or 'UINT32 'UNIX_FD)
                    (bytevector-u32-native-set! buffer 0 datum)
                    (values 4 4 #f)]
      ['INT64       (bytevector-s64-native-set! buffer 0 datum)
                    (values 8 8 #f)]
      ['UINT64      (bytevector-u64-native-set! buffer 0 datum)
                    (values 8 8 #f)]
      ['DOUBLE      (bytevector-ieee-double-native-set! buffer 0 datum)
                    (values 8 8 #f)]
      [(or 'STRING 'OBJECT_PATH)
       (let ((bv (string->utf8 (string-append datum "\x0;"))))
         (bytevector-u32-native-set! buffer 0 (fx- (bytevector-length bv) 1))
         (values 4 4 bv))]
      ['SIGNATURE
       (let ((bv (string->utf8 (string-append datum "\x0;"))))
         (bytevector-u8-set! buffer 0 (fx- (bytevector-length bv) 1))
         (values 1 1 bv))]

      ;; Compound types
      [('STRUCT . type*)
       (let-values ([(bvp extract) (open-bytevector-output-port)])
         (d-bus-put-message bvp 0 1 `(message ,@type*) datum)
         (values 0 8 (extract)))]
      [('DICT_ENTRY . type*)
       (match datum
         [(k . v)
          (let-values ([(bvp extract) (open-bytevector-output-port)])
            (d-bus-put-message bvp 0 1 `(message ,@type*) (list k v))
            (values 0 8 (extract)))]
         [_ (assertion-violation who "Invalid DICT_ENTRY data" sig)])]
      ['VARIANT
       (match datum
         [(signature . value)
          (let-values ([(bvp extract) (open-bytevector-output-port)])
            (d-bus-put-message bvp offset 1 `(message SIGNATURE ,signature)
                               (list (format-type-signature `(message ,signature))
                                     value))
            (values 0 1 (extract)))]
         [_ (assertion-violation who "Invalid VARIANT data" sig)])]
      [('ARRAY atype)
       (let-values ([(bvp extract) (open-bytevector-output-port)])
         (let ((type `(message ,atype)))
           (let lp ((i 0) (offset (fx+ 4 (fxalign offset 4))))
             (unless (= i (vector-length datum))
               (let ((offset^ (d-bus-put-message bvp offset 1 type
                                                 (list (vector-ref datum i)))))
                 (lp (fx+ i 1) offset^))))
           (let ((array-bv (extract)))
             (when (fx>? (bytevector-length array-bv) (expt 2 26))
               (assertion-violation who "Oversized array" sig))
             (bytevector-u32-native-set! buffer 0 (bytevector-length array-bv))
             (values 4 4 array-bv))))]
      [x (fail x)]))
  (unless (fx<=? 1 post-align 8)
    (assertion-violation who "Invalid post-align" port offset post-align sig data*))
  (match sig
    [('message type* ...)
     (let lp ((data* data*) (type* type*) (offset offset))
       (match type*
         [()
          (match data*
            [()
             (align-port port offset post-align)]
            [_ (assertion-violation who "Too much data" sig data*)])]
         [(type . rtype*)
          (match data*
            [() (assertion-violation who "Not enough data" sig data*)]
            [(datum . rdata*)
             (let-values ([(n alignment bv) (marshal-datum! buffer offset type datum)])
               (let ((offset (align-port port offset alignment)))
                 (put-bytevector port buffer 0 n)
                 (let ((n (cond ((bytevector? bv)
                                 (put-bytevector port bv)
                                 (fx+ n (bytevector-length bv)))
                                (else n))))
                   (lp rdata* rtype* (fx+ offset n)))))])]))]
    [x (fail x)])))
